__all__ = ['BlendCropper', 'Cropper', 'Decimator', 'Reconstructor', 'Zipper']

from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from math import ceil, floor

import numpy as np

from ._types import NumpyLike, Tile, Vec
from ._util import padslice


@dataclass(frozen=True, slots=True)
class Cropper:
    shape: tuple[int, ...]

    def __call__(self, tile: Tile) -> Tile:
        h, w = self.shape
        iyx, (y, x), data = tile
        return Tile(idx=iyx, vec=(y, x), data=data[:h - y, :w - x])


@dataclass(frozen=True, slots=True)
class Decimator:
    stride: int

    def __call__(self, tile: Tile) -> Tile:
        stride = self.stride
        vec = *(c // stride for c in tile.vec),
        data = tile.data[::stride, ::stride]
        return Tile(idx=tile.idx, vec=vec, data=data)  # type: ignore[arg-type]


@dataclass(frozen=True, slots=True)
class Zipper:
    v: NumpyLike
    v_scale: float

    def __call__(self, tile: Tile) -> tuple[Vec, Vec, np.ndarray, np.ndarray]:
        v_scale = self.v_scale
        loc = *(slice(floor(o * v_scale), ceil((o + s) * v_scale))
                for o, s in zip(tile.vec, tile.data.shape[:2])),
        r = padslice(self.v, *loc)
        return tile.idx, tile.vec, tile.data, r


class Reconstructor:
    """Joins non-overlapping parts to tiles"""
    def __init__(self, overlap: int, step: int, cells: np.ndarray):
        assert overlap
        self.overlap = overlap
        self.step = step
        self.cells = np.pad(cells, [(0, 1), (0, 1)])
        self.row = defaultdict[int, np.ndarray]()

    def __call__(self, t: Tile) -> Tile:
        (iy, ix), _, part = t

        # Lazy init, first part is always whole
        if self.row.default_factory is None:
            self.row.default_factory = partial(np.zeros, part.shape,
                                               part.dtype)

        if (tile := self.row.pop(ix, None)) is not None:
            tile[-part.shape[0]:, -part.shape[1]:] = part
        else:
            tile = part

        # Cache right & bottom edges for adjacent tiles
        if self.cells[iy, ix + 1]:
            self.row[ix + 1][:, :self.overlap] = tile[:, -self.overlap:]
        if self.cells[iy + 1, ix]:
            self.row[ix][:self.overlap, :] = tile[-self.overlap:, :]

        return Tile(
            idx=(iy, ix),
            vec=(iy * self.step - self.overlap, ix * self.step - self.overlap),
            data=tile,
        )


class BlendCropper:
    """
    Applies blends edges of incoming tiles.
    Returns crops excluding overlaps. Crop size depends on its surrounding.
    """
    def __init__(self, step: int, overlap: int, cells: np.ndarray):
        self.step = step
        self.overlap = overlap
        self.cells = np.pad(cells, [(0, 1), (0, 1)])
        self.row: dict[int, np.ndarray] = {}
        self.carry: np.ndarray | None = None

    def __call__(self, obj: Tile) -> Tile:
        """
        Blends edges of overlapping tiles and returns non-overlapping parts
        """
        (iy, ix), (y, x), tile = obj
        overlap = self.overlap
        step = self.step

        if iy and self.cells[iy - 1, ix]:  # TOP exists
            top = self.row.pop(ix)
            tile[:overlap, step - top.shape[1]:step] += top
        else:
            tile = tile[overlap:]  # cut TOP
            y += overlap

        if ix and self.cells[iy, ix - 1]:  # LEFT exists
            left = self.carry
            assert left is not None
            if self.cells[iy + 1, [ix - 1, ix]].all():
                tile[-left.shape[0]:, :overlap] += left
            else:  # cut BOTTOM-LEFT
                tile[-left.shape[0] - overlap:-overlap, :overlap] += left
        else:
            tile = tile[:, overlap:]  # cut LEFT
            x += overlap

        tile, right = np.split(tile, [-overlap], axis=1)
        if self.cells[iy, ix + 1]:  # RIGHT exists
            if not (iy and self.cells[iy - 1, [ix, ix + 1]].all()):
                right = right[-step:]  # cut TOP-RIGHT
            if not self.cells[iy + 1, [ix, ix + 1]].all():
                right = right[:-overlap]  # cut BOTTOM-RIGHT
            self.carry = right

        tile, bottom = np.split(tile, [-overlap])
        if self.cells[iy + 1, ix]:  # BOTTOM exists
            if not (ix and self.cells[[iy, iy + 1], ix - 1].all()):
                # cut BOTTOM-LEFT
                bottom = bottom[:, -(step - overlap):]
            self.row[ix] = bottom

        return Tile(idx=(iy, ix), vec=(y, x), data=tile)



__all__ = ['Slide']

import os
import warnings
from bisect import bisect_right
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass, field, replace
from math import ceil
from pathlib import Path
from typing import final, overload
from warnings import warn

import numpy as np
from glow import memoize, shared_call, weak_memoize

from bipl import env
from bipl.ops import normalize_loc, resize

from ._openslide import Openslide
from ._slide_bases import REGISTRY, Driver, Image, ImageLevel
from ._tiff import Tiff
from ._util import clahe, round2

# TODO: inside Slide.open import ._slide.registry,
# TODO: and in ._slide.registry do registration and DLL loading
# TODO: to make Slide export not require DLL presence

try:
    from ._gdal import Gdal
except ImportError:
    if not os.getenv('_BIPL_GDAL_NO_WARN'):
        msg = 'No GDAL is available. Please '
        if os.name == 'nt':
            msg += ('acquire it manually from '
                    'https://github.com/cgohlke/geospatial-wheels/releases ')
        else:
            msg += ('install libgdal via your system package manager, '
                    'and run "pip install gdal==`gdal-config --version`"')
        warn(msg, stacklevel=1)
        os.environ['_BIPL_GDAL_NO_WARN'] = '1'
    Gdal = None  # type: ignore[assignment,misc]

_drv: type[Driver] | None
for _drv, _regex in [  # LIFO, last driver takes priority
    (Gdal, r'^.*\.(tif|tiff)$'),
    (Openslide, r'^.*\.(bif|mrxs|ndpi|scn|svs|svsslide|tif|tiff|vms|vmu)$'),
    (Tiff, r'^.*\.(svs|tif|tiff)$'),
    (Gdal, r'^(/vsicurl.*|(http|https)://.*)$'),
]:
    if _drv is not None and _drv.__name__.lower() in env.BIPL_DRIVERS:
        _drv.register(_regex)


@shared_call  # merge duplicate calls
@weak_memoize  # reuse result if it's already exist, but used by someone else
@memoize(capacity=env.BIPL_CACHE, policy='lru')  # keep LRU for unused results
def _cached_open(path: str, **kwargs) -> 'Slide':
    last_exc = BaseException()
    matches = (tp for pat, tps_ in REGISTRY.items() if pat.match(path)
               for tp in tps_)
    tps = [*dict.fromkeys(matches)]
    if tps:
        for tp in reversed(tps):  # Loop over types to find non-failing
            try:
                return Slide.from_file(path, tp).tonemap(**kwargs)
            except (ValueError, TypeError) as exc:
                last_exc = exc
        raise last_exc from None
    raise ValueError(f'Unknown file format {path}')


@final
@dataclass(frozen=True)
class Slide:
    """Usage:
    ```
    slide = Slide.open('test.svs')
    shape: tuple[int, ...] = slide.shape
    scales: tuple[int, ...] = slide.scales

    # Get numpy.ndarray
    image: np.ndarray = slide[:2048, :2048]
    ```
    """
    # TODO: check if memory leak
    # TODO: add __enter__/__exit__/close to make memory management explicit
    # TODO: call .close in finalizer
    path: str
    shape: Sequence[int]
    downsamples: Sequence[int]
    mpp: float | None
    driver: str
    bbox: tuple = field(repr=False)
    levels: Sequence[ImageLevel] = field(repr=False)
    extras: Mapping[str, Image] = field(repr=False)

    @classmethod
    def from_file(
        cls,
        path: str,
        driver_fn: Callable[[str], Driver],
        mpp: float | None = None,
    ) -> 'Slide':
        driver = driver_fn(path)

        num_images = len(driver)
        if num_images == 0:
            raise ValueError('Empty file')

        im_0, *images = (driver[idx] for idx in range(num_images))
        if not isinstance(im_0, ImageLevel):
            raise TypeError('First pyramid layer is not tiled')

        # Retrieve all sub-images
        levels: dict[int, ImageLevel] = {1: im_0}
        extras: dict[str, Image] = {}
        for im in images:
            if isinstance(im, ImageLevel):
                downsample = round2(im_0.shape[0] / im.shape[0])
                levels[downsample] = im
            elif key := im.key:
                extras[key] = im
        extras |= driver.named_items()

        level_downsamples = *sorted(levels.keys()),
        # TODO: make virtual levels if downsamples are too distant (ProxyLevel)

        if mpp is None:  # If no override is passed, use native if present
            mpp = levels[1].mpp

        return Slide(
            path=path,
            shape=levels[1].shape,
            mpp=mpp,
            downsamples=level_downsamples,
            driver=type(driver).__name__,
            bbox=driver.bbox,
            levels=tuple(levels[zoom] for zoom in level_downsamples),
            extras=extras,
        )

    def icc(self) -> 'Slide':
        if icc_0 := self.levels[0].icc:
            levels = *(il.apply(icc_0) for il in self.levels),
            self = replace(self, levels=levels)

        if any(e.icc for e in self.extras.values()):
            extras = {
                t: e.apply(e.icc) if e.icc else e
                for t, e in self.extras.items()
            }
            self = replace(self, extras=extras)

        return self

    def clahe(self) -> 'Slide':
        levels = *(il.apply(clahe, pad=64) for il in self.levels),
        extras = {k: i.apply(clahe) for k, i in self.extras.items()}
        return replace(self, levels=levels, extras=extras)

    def tonemap(self, icc: bool, clahe: bool) -> 'Slide':
        r = self
        if icc:
            r = r.icc()
        if clahe:
            r = r.clahe()
        return r

    def mpp_or_error(self) -> float:
        if self.mpp is None:
            raise ValueError('Slide`s MPP is unknown')
        return self.mpp

    @property
    def pools(self) -> Sequence[int]:
        warnings.warn(
            '"Slide.pools" is deprecated. Use "Slide.downsamples"',
            category=DeprecationWarning,
            stacklevel=2)
        return self.downsamples

    @property
    def spacing(self) -> float | None:
        warnings.warn(
            '"Slide.spacing" is deprecated. Use "Slide.mpp"',
            category=DeprecationWarning,
            stacklevel=2)
        return self.mpp

    def __reduce__(self) -> tuple:
        return Slide.open, (self.path, )

    def best_level_for(
        self,
        downsample: float,
        /,
        *,
        tol: float = 0.01,
    ) -> tuple[int, ImageLevel]:
        """Gives the most detailed LOD below `downsample`"""
        downsample = downsample * max(1, 1 + tol)
        idx = max(bisect_right(self.downsamples, downsample) - 1, 0)
        return self.downsamples[idx], self.levels[idx]

    def resample(self, mpp: float, *, tol: float = 0.01) -> ImageLevel:
        """Resample slide to specific resolution"""
        downsample = mpp / self.mpp_or_error()
        return self.pool(downsample, tol=tol)

    def pool(self, downsample: float, /, *, tol: float = 0.01) -> ImageLevel:
        """Use like `slide.pool(4)[y0:y1, x0:x1]` call"""
        d, lvl = self.best_level_for(downsample, tol=tol)
        if d == downsample:
            return lvl
        return lvl.rescale(d / downsample)

    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        """Retrieves tile"""
        # TODO: Ignore step, always redirect to self.levels[0].__getitem__
        y_loc, x_loc, c_loc = normalize_loc(key, self.shape)

        try:
            step, = {y_loc.step, x_loc.step}
        except ValueError:
            raise ValueError('all slices should have the same step') from None
        if step <= 0:
            raise ValueError('slice steps should be positive')

        ds, level = self.best_level_for(step)
        yx_loc = *(slice(s.start // ds, s.stop // ds) for s in (y_loc, x_loc)),
        image = level.crop(*yx_loc)

        dsize = *(ceil((s.stop - s.start) / step) for s in (y_loc, x_loc)),
        return resize(image, dsize)[:, :, c_loc]

    @overload
    def at(self,
           z0_yx_offset: tuple[int, ...],
           dsize: int | tuple[int, ...],
           *,
           scale: float,
           tol: float = ...) -> np.ndarray:
        ...

    @overload
    def at(self,
           z0_yx_offset: tuple[int, ...],
           dsize: int | tuple[int, ...],
           *,
           mpp: float,
           tol: float = ...) -> np.ndarray:
        ...

    def at(self,
           z0_yx_offset: tuple[int, ...],
           dsize: int | tuple[int, ...],
           *,
           scale: float | None = None,
           mpp: float | None = None,
           tol: float = 0.01) -> np.ndarray:
        """Read square region starting with offset"""
        dsize = dsize if isinstance(dsize, tuple) else (dsize, dsize)
        if len(dsize) != 2:
            raise ValueError(f'dsize should be 2-tuple or int. Got {dsize}')

        if scale is None:
            if mpp is None:
                raise ValueError('Only one of zoom/mpp should be None')
            scale = self.mpp_or_error() / mpp

        ds, lvl = self.best_level_for(1 / scale, tol=tol)
        loc = *(slice(int(c / ds), int((c + size / scale) / ds))
                for c, size in zip(z0_yx_offset, dsize)),
        image = lvl.crop(*loc)
        return resize(image, dsize)

    def extra(self, name: str) -> np.ndarray | None:
        if im := self.extras.get(name):
            return im.numpy()
        return None

    def thumbnail(self) -> np.ndarray:
        return self.extras.get('thumbnail', self.levels[-1]).numpy()

    @classmethod
    def open(cls,
             anypath: Path | str,
             /,
             *,
             icc: bool = env.BIPL_ICC,
             clahe: bool = env.BIPL_CLAHE) -> 'Slide':
        """Open multi-scale image."""
        if isinstance(anypath, Path):
            anypath = Path(anypath).resolve().absolute().as_posix()
        return _cached_open(anypath, icc=icc, clahe=clahe)



__all__ = ['Dzi']

from typing import TYPE_CHECKING, Literal, NamedTuple

import cv2
import numpy as np
from lxml.builder import ElementMaker
from lxml.etree import tostring

if TYPE_CHECKING:
    from ._slide import Slide


class Dzi(NamedTuple):
    tile: int
    quality: int = 90
    fmt: Literal['jpeg', 'webp', 'avif'] = 'webp'

    def head(self, slide: 'Slide') -> str:
        e = ElementMaker()
        return tostring(
            e.Image(
                e.Size(Height=str(slide.shape[0]), Width=str(slide.shape[1])),
                TileSize=str(self.tile),
                Overlap='0',
                Format=self.fmt,
                xmlns='http://schemas.microsoft.com/deepzoom/2008'))

    def tile_at(self, slide: 'Slide', level: int,
                iy_ix: tuple[int, ...]) -> np.ndarray:
        scale = 2 ** (level - max(slide.shape[:2]).bit_length())
        tile_0 = self.tile * scale
        offset_0 = *(ip * tile_0 for ip in iy_ix),
        return slide.at(offset_0, self.tile, scale=scale)

    def compress(self, rgb: np.ndarray) -> bytes:
        qtag = {
            'jpeg': cv2.IMWRITE_JPEG_QUALITY,
            'webp': cv2.IMWRITE_WEBP_QUALITY,
        }.get(self.fmt)
        if qtag is None:
            raise RuntimeError(f'Unknown format: {self.fmt}')

        bgr = cv2.cvtColor(rgb, cv2.COLOR_RGB2BGR)
        _, data = cv2.imencode(f'.{self.fmt}', bgr, [qtag, self.quality])
        return data.tobytes()



__all__ = ['Driver', 'Image', 'ImageLevel', 'REGISTRY']

import re
from collections.abc import Callable
from dataclasses import dataclass
from math import ceil
from typing import TYPE_CHECKING, final

import cv2
import numpy as np

from bipl import env
from bipl.ops import Tile, get_fusion, normalize_loc, resize

if TYPE_CHECKING:
    from ._util import Icc

REGISTRY: dict[re.Pattern, list[type['Driver']]] = {}
_MIN_TILE = 256


@dataclass(frozen=True)
class Image:
    shape: tuple[int, ...]

    @property
    def key(self) -> str | None:
        raise NotImplementedError

    def numpy(self) -> np.ndarray:
        """Convert to ndarray"""
        raise NotImplementedError

    @final
    def __array__(self) -> np.ndarray:
        """numpy.array compatibility"""
        return self.numpy()

    def apply(self, fn: Callable[[np.ndarray], np.ndarray]) -> '_LambdaImage':
        return _LambdaImage(self.shape, self, fn)

    @property
    def icc(self) -> 'Icc | None':
        return None


@dataclass(frozen=True)
class ImageLevel(Image):
    mpp: float | None

    @final
    @property
    def key(self) -> None:
        return None

    def crop(self, *loc: slice) -> np.ndarray:
        """Reads crop of LOD. Overridable"""
        raise NotImplementedError

    @final
    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        """Reads crop of LOD"""
        y_loc, x_loc, c_loc = normalize_loc(key, self.shape)
        if not y_loc.step == x_loc.step == 1:
            raise ValueError('Y/X slice steps should be 1 for now, '
                             f'got {y_loc.step} and {x_loc.step}')
        return self.crop(y_loc, x_loc)[:, :, c_loc]

    @final
    def numpy(self) -> np.ndarray:
        """Reads whole image in single op"""
        return self[:, :]

    def rescale(self, scale: float) -> 'ImageLevel':
        """
        Resize image to `src.size * scale`.
        I.e. downscale if `scale < 1`, upscale otherwise.
        """
        if scale == 1:
            return self

        base = self
        if isinstance(base, ProxyLevel):
            scale = base.scale * scale
            base = base.base

        h, w, c = base.shape
        h, w = (round(h * scale), round(w * scale))  # TODO: round/ceil/floor ?
        mpp = base.mpp / scale if base.mpp else None
        if scale > 0.5:  # Downscale to less then 2x, or upsample
            return ProxyLevel((h, w, c), mpp, scale, base)

        downsample = 2 ** (int(1 / scale).bit_length() - 1)
        r_tile = max(_MIN_TILE // downsample, 1)
        return TiledProxyLevel((h, w, c), mpp, scale, base, downsample, r_tile)

    def _unpack_loc(
        self,
        *slices: slice,
    ) -> tuple[np.ndarray, np.ndarray, list[int]]:
        box = np.array([(s.start, s.stop) for s in slices])
        valid_box = box.T.clip([0, 0], self.shape[:2]).T  # (2, lo-hi)
        shape = (box[:, 1] - box[:, 0]).tolist()
        return box, valid_box, shape

    def _expand(self, rgb: np.ndarray, valid_box: np.ndarray, box: np.ndarray,
                bg_color: np.ndarray) -> np.ndarray:
        offsets = np.abs(valid_box - box)
        if offsets.any():
            tp, bm, lt, rt = offsets.ravel().tolist()
            rgb = cv2.copyMakeBorder(rgb, tp, bm, lt, rt, cv2.BORDER_CONSTANT,
                                     None, bg_color.tolist())
        return np.ascontiguousarray(rgb)

    @final
    def apply(  # type: ignore[override]
            self,
            fn: Callable[[np.ndarray], np.ndarray],
            pad: int = 0) -> '_LambdaLevel':
        # _LambdaLevel is not subclass of _LambdaImage
        return _LambdaLevel(self.shape, self.mpp, self, fn, pad)


@dataclass(frozen=True)
class _LambdaImage(Image):
    base: Image
    fn: Callable[[np.ndarray], np.ndarray]

    def numpy(self) -> np.ndarray:
        im = self.base.numpy()
        return self.fn(im)


@dataclass(frozen=True)
class _LambdaLevel(ImageLevel):
    base: ImageLevel
    fn: Callable[[np.ndarray], np.ndarray]
    pad: int = 64

    def crop(self, *loc: slice) -> np.ndarray:
        loc_ = *(slice(s.start - self.pad, s.stop + self.pad, s.step)
                 for s in loc),
        im = self.base.crop(*loc_)
        im = self.fn(im)
        if not self.pad:
            return im
        return im[self.pad:-self.pad, self.pad:-self.pad, :]


@dataclass(frozen=True)
class ProxyLevel(ImageLevel):
    scale: float
    base: ImageLevel

    def _get_loc(self, *src_loc: slice) -> np.ndarray:
        return self.base[src_loc]

    def crop(self, *loc: slice) -> np.ndarray:
        src_loc = *[
            # TODO: round/ceil/floor ?
            slice(round(s.start / self.scale), round(s.stop / self.scale))
            for s in loc
        ],
        image = self._get_loc(*src_loc)

        h, w = ((s.stop - s.start) for s in loc)
        return resize(image, (h, w))


@dataclass(frozen=True)
class TiledProxyLevel(ProxyLevel):
    downsample: int
    r_tile: int

    def _get_loc(self, *src_loc: slice) -> np.ndarray:
        s_start = [s.start for s in src_loc]
        s_shape = *(s.stop - s.start for s in src_loc),
        if np.prod(s_shape) < env.BIPL_TILE_POOL_SIZE:
            return super()._get_loc(*src_loc)

        r_shape = *(ceil(size / self.downsample) for size in s_shape),
        r_tile = self.r_tile
        s_tile = r_tile * self.downsample

        ty, tx = (ceil(size / s_tile) for size in s_shape)
        tgrid = np.mgrid[:ty, :tx].reshape(2, -1).T

        t_shape = (r_tile, r_tile)
        r_tiles = map(
            Tile,
            tgrid.tolist(),
            (tgrid * r_tile).tolist(),
            (resize(self.base[sy:sy + s_tile, sx:sx + s_tile], t_shape)
             for sy, sx in (tgrid * s_tile + s_start).tolist()),
        )
        image = get_fusion(r_tiles, r_shape)
        assert image is not None
        return image


class Driver:
    @final
    @classmethod
    def register(cls, regex: str):
        """Registers type builder for extensions. Last call takes precedence"""
        REGISTRY.setdefault(re.compile(regex), []).append(cls)

    def __init__(self, path: str) -> None:
        raise NotImplementedError

    def __len__(self) -> int:
        """Count of indexed images, usually resolution images"""
        return 0

    def __getitem__(self, index: int) -> Image:
        """Gives indexed image"""
        raise NotImplementedError

    def named_items(self) -> dict[str, Image]:
        keys = self.keys()
        return {k: self.get(k) for k in keys}

    def keys(self) -> list[str]:
        """Names of associated images"""
        return []

    def get(self, key: str) -> Image:
        """Get assosiated image by key"""
        raise NotImplementedError

    @property
    def bbox(self) -> tuple[slice, slice]:
        return slice(None), slice(None)



__all__ = ['env']

from typing import Literal

from pydantic import BaseSettings, ByteSize, parse_obj_as


class Env(BaseSettings):
    BIPL_DRIVERS: set[str] = {'gdal', 'tiff', 'openslide'}
    BIPL_CACHE: ByteSize = parse_obj_as(ByteSize, '10 MiB')
    BIPL_TILE_CACHE: int = 100  # Max tiles cached per tiff slide

    # area - downsample in single `resize(inter=area)` call (slow)
    # box2d - if downsample more than 2X, do gradual `resize(f=0.5)` first
    # gauss - do pyramid downsampling (2x slow than box2d), no aliasing
    BIPL_DOWN: Literal['area', 'box2x', 'gauss'] = 'gauss'  # Downsampling mode

    BIPL_TILE_POOL_SIZE: int = 64_000_000  # Min resolution for tiled pooling
    BIPL_ICC: bool = False  # Apply image color correction
    BIPL_CLAHE: bool = False


env = Env()



__all__ = ['NumpyLike', 'Tile', 'Vec']

from typing import NamedTuple, Protocol

import numpy as np

Vec = tuple[int, int]


class NumpyLike(Protocol):
    @property
    def shape(self) -> tuple[int, ...]:
        ...

    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        ...


class Tile(NamedTuple):
    idx: Vec
    vec: Vec
    data: np.ndarray



"""
Driver based on libtiff
- fast
- not thread safe (internally)
- compatible with TIFF and its flavours
"""

__all__ = ['Tiff']

import ctypes
import sys
import weakref
from collections.abc import Iterator
from contextlib import contextmanager
from ctypes import (POINTER, addressof, byref, c_char_p, c_float, c_int,
                    c_ubyte, c_uint16, c_uint32, c_uint64, c_void_p,
                    create_string_buffer, string_at)
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from threading import Lock
from typing import NamedTuple, Protocol, TypeVar

import cv2
import imagecodecs
import numpy as np

from bipl._env import env

from ._libs import load_library
from ._slide_bases import Driver, Image, ImageLevel
from ._util import Icc, is_aperio, parse_aperio_description

_T = TypeVar('_T')

TIFF = load_library('libtiff', 6, 5)
# _TIFF.TIFFSetErrorHandler(None)

(TIFF.TIFFOpenW if sys.platform == 'win32' else TIFF.TIFFOpen).restype \
    = POINTER(c_ubyte)


# ---------------------------- TIFF tags mappings ----------------------------
class _ColorSpace(Enum):
    MINISBLACK = 1
    RGB = 2
    YCBCR = 6


class _ColorInfo(NamedTuple):
    space: _ColorSpace
    subsampling: tuple[int, int]

    def to_rgb(self, x: np.ndarray) -> np.ndarray:
        if self.space is not _ColorSpace.YCBCR or self.subsampling != (2, 2):
            return x

        h, w = x.shape[:2]
        assert h % 2 == w % 2 == 0

        # (2 h/2) w 3 -> h/2 w 3
        # h/2 (w/2 2) 3 -> h/2 w/2 2 3 -> h/2 w/2 (2 3) -> h/2 w/2 6
        hw6 = x[:h // 2, :, :].reshape(h // 2, w // 2, 6)

        # y00, y01, y10, y11, cb, cr
        y = hw6[:, :, [[0, 1], [2, 3]]].transpose(0, 2, 1, 3).reshape(h, w)
        cb_cr = hw6[:, :, [4, 5]]
        return cv2.cvtColorTwoPlane(y, cb_cr, cv2.COLOR_YUV2RGB_NV12)


class _Compression(Enum):
    RAW = 1
    CCITT = 2
    CCITTFAX3 = 3
    CCITTFAX4 = 4
    LZW = 5
    JPEG_OLD = 6
    JPEG = 7
    ADOBE_DEFLATE = 8
    RAW_16 = 32771
    PACKBITS = 32773
    THUNDERSCAN = 32809
    DEFLATE = 32946
    DCS = 32947
    JPEG2000_YUV = 33003
    JPEG2000_RGB = 33005
    JBIG = 34661
    SGILOG = 34676
    SGILOG24 = 34677
    JPEG2000 = 34712
    LZMA = 34925
    ZSTD = 50000
    WEBP = 50001


class _Tag:  # noqa: PIE795,RUF100
    JPEG_TABLES = 347
    TILE_BYTE_COUNTS = 325
    BACKGROUND_COLOR = 434


class _Tags:
    def __init__(self, ptr):
        self._ptr = ptr
        compression, = self._get(c_uint16, 259)
        self.compression = _Compression(compression)

        self.spp, = self._get(c_uint16, 277)
        self.is_planar, = self._get(c_uint16, 284)
        self.image_size = self._get(c_uint32, 257, 256)
        self.tile_size = self._get(c_uint32, 323, 322)
        self.description = self._get_str(270)
        self.resolution = self._get(c_float, 283, 282)

        # ! crashes, but should work according to openslide docs
        # self.is_hamamatsu = bool(self._get(c_uint16, 65420))
        self.make = self._get_str(271)

        self.bps, = self._get(c_uint16, 339) or [1]

        ics, = self._get(c_uint16, 262)
        colorspace = _ColorSpace(ics)
        subsampling = (1, 1)

        # TODO: use this in YCbCr conversion
        self.gray = np.array([], 'f4')  # Luma coefficients
        self.yuv_centered = True
        self.yuv_bw = np.array([], 'f4')  # BW pairs, per channel

        if colorspace is _ColorSpace.YCBCR:
            self.gray = np.array([.299, .587, .114], 'f4')
            gray_ptr = POINTER(c_float)()
            if TIFF.TIFFGetField(self._ptr, 529, byref(gray_ptr)):
                self.gray = np.ctypeslib.as_array(gray_ptr, [3]).copy()

            # YCbCr subsampling H/W, i.e:
            # 24bps:
            #  (1 1) = 4:4:4 (cccc/cccc), 100% H, 100% W
            # 16bps:
            #  (1 2) = 4:2:2 (c_c_/c_c_), 100% H, 50% W
            #  (2 1) = 4:4:0 (cccc/____), 50% H, 100% W
            # 12bps:
            #  (1 4) = 4:1:1 (c___/c___), 100% H, 25% W
            #  (2 2) = 4:2:0 (c_c_/____), 50% H, 50% W - a.k.a. YUV-NV12
            # 10bps:
            #  (2 4) = 4:1:0 (c___/____), 50% H, 25% W
            ss_w, ss_h = self._get_varargs((c_uint16, c_uint16), 530) or (2, 2)
            subsampling = ss_h, ss_w

            self.yuv_centered = 2 not in self._get(c_uint16, 531)

            bw_ptr = POINTER(c_float)()
            if TIFF.TIFFGetField(self._ptr, 532, byref(bw_ptr)):
                self.yuv_bw = np.ctypeslib.as_array(bw_ptr, [3, 2])

        self.color = _ColorInfo(colorspace, subsampling)

        self.icc = None
        icc_size = c_int()
        icc_ptr = c_char_p()
        if (TIFF.TIFFGetField(self._ptr, 34675, byref(icc_size),
                              byref(icc_ptr)) and icc_size.value > 0):
            self.icc = Icc(string_at(icc_ptr, icc_size.value))

    def _get(
        self,
        tp: type['ctypes._SimpleCData[_T]'],
        *tags: int,
    ) -> tuple[_T, ...]:
        values: list[_T] = []
        for tag in tags:
            cv = tp()
            if TIFF.TIFFGetField(self._ptr, c_uint32(tag), byref(cv)):
                values.append(cv.value)
        return *values,

    def _get_str(self, tag: int) -> str:
        values = self._get(c_char_p, tag)
        if not values:
            return ''
        return (values[0] or b'').decode()

    def _get_varargs(self, tps: tuple[type['ctypes._SimpleCData[_T]'], ...],
                     tag: int) -> tuple[_T, ...]:
        cvs = *(tp() for tp in tps),
        if TIFF.TIFFGetField(self._ptr, c_uint32(tag), *map(byref, cvs)):
            return *(cv.value for cv in cvs),
        return ()


# -------------------------- lazy decoding proxies ---------------------------
class SupportsArray(Protocol):
    def __array__(self) -> np.ndarray:
        raise NotImplementedError


class _CachedArray:
    def __array__(self) -> np.ndarray:
        return self.numpy

    @cached_property
    def numpy(self) -> np.ndarray:
        return self._impl()

    def _impl(self) -> np.ndarray:
        raise NotImplementedError


@dataclass
class ImageArray(_CachedArray):
    data: object

    def _impl(self) -> np.ndarray:
        return imagecodecs.imread(self.data)


@dataclass
class JpegArray(_CachedArray):
    data: object
    jpt: bytes
    colorspace: int

    def _impl(self) -> np.ndarray:
        return imagecodecs.jpeg_decode(
            self.data, tables=self.jpt, colorspace=self.colorspace)


# ------------------ image, level & opener implementations -------------------


@dataclass(frozen=True)
class _BaseImage(Image):
    index: int
    icc_impl: Icc | None
    tiff: 'Tiff'

    @property
    def icc(self) -> Icc | None:
        return self.icc_impl


@dataclass(frozen=True)
class _Image(_BaseImage):
    head: list[str]
    vendor: str

    @property
    def key(self) -> str | None:
        if self.vendor == 'aperio' and self.index == 1:
            return 'thumbnail'
        for key in ('label', 'macro'):
            if any(key in s for s in self.head):
                return key
        return None

    def numpy(self) -> np.ndarray:
        h, w = self.shape[:2]
        rgba = np.empty((h, w, 4), dtype='u1')

        with self.tiff.ifd(self.index) as ptr:
            ok = TIFF.TIFFReadRGBAImageOriented(ptr, w, h,
                                                c_void_p(rgba.ctypes.data), 1,
                                                0)
            if not ok:
                raise ValueError('TIFF image read failed')

        rgba = cv2.cvtColor(rgba, cv2.COLOR_mRGBA2RGBA)
        # TODO: do we need to use bg_color to fill points where alpha = 0 ?
        return rgba[..., :3]


@dataclass(frozen=True)
class _Level(ImageLevel, _BaseImage):
    color: _ColorInfo
    bg_color: np.ndarray
    compression: _Compression
    jpt: bytes = field(repr=False)
    tile: tuple[int, ...]
    tile_sizes: np.ndarray = field(repr=False)

    def _read_tile(self, y, x, ptr) -> SupportsArray:
        offset = TIFF.TIFFComputeTile(ptr, x, y, 0, 0)
        nbytes = int(self.tile_sizes[offset])

        if not nbytes:  # If nothing to read, don't read
            raise ValueError('File has corrupted tiles with zero size')
            # TODO: read from previous level
            # * If tile is empty on level N,
            # * then all tiles on levels >N are invalid, whether empty or not
            return np.broadcast_to(self.bg_color, self.tile)

        # if not self.compression.name.startswith('JPEG2000'):
        if self.compression not in {
                _Compression.JPEG2000_RGB, _Compression.JPEG2000_YUV
        }:
            image = np.empty(self.tile, dtype='u1')
            isok = TIFF.TIFFReadTile(ptr, c_void_p(image.ctypes.data), x, y, 0,
                                     0)
            if isok == -1:
                raise ValueError('TIFF tile read failed')
            return self.color.to_rgb(image)

        data = create_string_buffer(nbytes)
        TIFF.TIFFReadRawTile(ptr, offset, data, len(data))

        if self.jpt:
            return JpegArray(data, self.jpt, self.color.space.value)
        return ImageArray(data)

    def _get_tile(self, y: int, x: int, ptr, cache_ok: bool) -> SupportsArray:
        # Runs within lock
        key = (self.index, y, x)
        cache = self.tiff.cache
        if (obj := cache.pop(key, None)) is not None:
            # Cache hit, move to the end
            cache[key] = obj
        else:
            # Cache miss
            obj = self._read_tile(y, x, ptr)

            if cache_ok and (cache_cap := env.BIPL_TILE_CACHE):
                while len(cache) >= cache_cap:  # Evict least used item
                    cache.pop(next(iter(cache)))
                cache[key] = obj
        return obj

    def crop(self, *loc: slice) -> np.ndarray:
        box = np.array([(s.start, s.stop) for s in loc])

        *tile, spp = self.tile
        dyx = box[:, 0]  # (2 lo-hi) -> (2)
        out = np.ascontiguousarray(
            np.broadcast_to(
                self.bg_color,
                np.r_[box[:, 1] - box[:, 0], spp],
            ))

        bmin, bmax = box.T.clip(0, self.shape[:2])

        axes = *map(slice, bmin // tile, -(-bmax // tile)),
        t_lo = np.mgrid[axes].reshape(2, -1).T * tile  # [N, 2]
        if not t_lo.size:
            return out

        # Cache only edges
        # TODO: profile whether it gives any perf benefits
        is_edge = (t_lo == t_lo.min(0)).any(1) | (t_lo == t_lo.max(0)).any(1)
        with self.tiff.ifd(self.index) as ptr:
            parts = [
                self._get_tile(y, x, ptr, cache_ok)
                for (y, x), cache_ok in zip(t_lo.tolist(), is_edge.tolist())
            ]

        # [N, lo-hi, yx]
        crops = np.stack([t_lo, t_lo + tile], 1).clip(bmin, bmax)

        # [N, yx, lo-hi]
        o_crops = (crops - dyx).transpose(0, 2, 1)
        t_crops = (crops - t_lo[:, None, :]).transpose(0, 2, 1)
        for part, (oy, ox), (ty, tx) in zip(parts, o_crops, t_crops):
            patch = np.array(part, copy=False)
            out[slice(*oy), slice(*ox)] = patch[slice(*ty), slice(*tx)]

        return out


# FIXME: Get around slides from choked SVS encoder
class Tiff(Driver):
    def __init__(self, path: str):
        # TODO: use memmap instead of libtiff
        self._ptr = (
            TIFF.TIFFOpenW(path, b'rm') if sys.platform == 'win32' else
            TIFF.TIFFOpen(path.encode(), b'rm'))
        if not self._ptr:
            raise ValueError(f'File {path} cannot be opened')

        weakref.finalize(self, TIFF.TIFFClose, self._ptr)

        self._lock = Lock()
        self.cache: dict[tuple, SupportsArray] = {}

    def __repr__(self) -> str:
        return f'{type(self).__name__}({addressof(self._ptr.contents):0x})'

    @contextmanager
    def ifd(self, index: int) -> Iterator:
        with self._lock:
            TIFF.TIFFSetDirectory(self._ptr, index)
            try:
                yield self._ptr
            finally:
                TIFF.TIFFFreeDirectory(self._ptr, index)

    def __len__(self) -> int:
        return TIFF.TIFFNumberOfDirectories(self._ptr)

    def _bg_color(self) -> np.ndarray:
        bg_hex = b'FFFFFF'
        bg_color_ptr = c_char_p()
        # TODO: find sample file to test this path. Never reached
        if TIFF.TIFFGetField(self._ptr, _Tag.BACKGROUND_COLOR,
                             byref(bg_color_ptr)):
            bg_hex = string_at(bg_color_ptr, 3)
            # TIFF._TIFFfree(bg_color_ptr)  # TODO: ensure no segfault
        return np.frombuffer(bytes.fromhex(bg_hex.decode()), 'u1')

    def _parse_description(self, desc: str,
                           make: str) -> tuple[str, list[str], dict[str, str]]:
        vendor = ''
        if is_aperio(desc):
            vendor = 'aperio'
            head, meta = parse_aperio_description(desc)

        else:
            if make == 'Hamamatsu':
                raise ValueError('Hamamatsu is not yet supported via libtiff')
            # TODO: put xml parser here (tiff)
            head = [desc]
            meta = {}

        return vendor, head, meta

    def _mpp(self, resolution: tuple[float, ...],
             meta: dict[str, str]) -> float | None:
        if s := [(10_000 / v) for v in resolution if v]:
            return float(np.mean(s))
        if mpp := meta.get('MPP'):
            return float(mpp)
        return None

    def _get(self, index: int) -> Image:
        tags = _Tags(self._ptr)

        if tags.is_planar != 1:
            raise TypeError(f'Level {index} is not contiguous!')

        bg_color = self._bg_color()
        vendor, head, meta = self._parse_description(tags.description,
                                                     tags.make)
        mpp = self._mpp(tags.resolution, meta) if index == 0 else None

        if (tags.color.space is _ColorSpace.YCBCR
                and tags.color.subsampling != (2, 2)):
            raise ValueError('Unsupported YUV subsampling: '
                             f'{tags.color.subsampling}')

        # Compression and JPEG tables
        jpt = b''
        if tags.compression is _Compression.JPEG:
            count = c_int()
            jpt_ptr = c_char_p()
            if TIFF.TIFFGetField(self._ptr, _Tag.JPEG_TABLES, byref(count),
                                 byref(jpt_ptr)) and count.value > 4:
                jpt = string_at(jpt_ptr, count.value)
                # TIFF._TIFFfree(jpt_ptr)  # TODO: ensure no segfault

        # Whole level shape
        shape = (*tags.image_size, tags.spp)
        if len(shape) != 3:
            raise ValueError(f'Bad shape in TIFF: {shape}')

        # Tile shape, if applicable
        if not TIFF.TIFFIsTiled(self._ptr):  # Not yet supported
            return _Image(shape, index, tags.icc, self, head, vendor)

        # Tile sizes
        tile = (*tags.tile_size, tags.spp)
        if len(tile) != 3:
            raise ValueError(f'Bad tile shape in TIFF: {tile}')

        tbc = np.empty([], 'u8')
        tbc_ptr = POINTER(c_uint64)()
        if TIFF.TIFFGetField(self._ptr, _Tag.TILE_BYTE_COUNTS, byref(tbc_ptr)):
            num_tiles = TIFF.TIFFNumberOfTiles(self._ptr)
            tbc = np.ctypeslib.as_array(tbc_ptr, [num_tiles]).copy()
            # TIFF._TIFFfree(tbc_ptr)  # TODO: ensure no segfault
            if (tbc == 0).any():
                raise ValueError('Found 0s in tile size table')

        return _Level(shape, index, tags.icc, self, mpp, tags.color, bg_color,
                      tags.compression, jpt, tile, tbc)

    def __getitem__(self, index: int) -> Image:
        with self.ifd(index):
            return self._get(index)



from ._mosaic import Mosaic
from ._types import Tile
from ._util import (get_fusion, get_trapz, normalize_loc, probs_to_rgb_heatmap,
                    resize)

__all__ = [
    'Mosaic',
    'Tile',
    'get_fusion',
    'get_trapz',
    'normalize_loc',
    'probs_to_rgb_heatmap',
    'resize',
]



__all__ = ['Mosaic']

import dataclasses
from collections.abc import Callable, Iterable, Iterator
from contextlib import AbstractContextManager
from dataclasses import dataclass
from functools import partial
from itertools import chain
from math import ceil
from typing import TypeVar, cast

import cv2
import numpy as np
from glow import chunked, map_n

from ._tile import BlendCropper, Cropper, Decimator, Reconstructor, Zipper
from ._types import NumpyLike, Tile, Vec
from ._util import get_trapz, padslice

# TODO: allow result of .map/.map_batched to have different tile and step
# NOTE: all classes here are stateless & their iterators are reentrant.

# ---------------------------- utility functions -----------------------------


def _apply(fn: Callable[[np.ndarray], np.ndarray], obj: Tile) -> Tile:
    r = fn(obj.data)
    assert r.shape[:2] == obj.data.shape[:2], \
        'Tile shape alteration (besides channel count) is forbidden'
    return obj._replace(data=r)


def _apply_batched(fn: Callable[[list[np.ndarray]], Iterable[np.ndarray]],
                   ts: tuple[Tile, ...]) -> list[Tile]:
    *rs, = fn([t.data for t in ts])
    assert len(rs) == len(ts)
    assert all(r.shape[:2] == t.data.shape[:2]
               for r, t in zip(rs, ts)), \
        'Tile shape alteration (besides channel count) is forbidden'
    return [t._replace(data=r) for t, r in zip(ts, rs)]


def _reweight(weight: np.ndarray, tile: np.ndarray) -> np.ndarray:
    assert tile.dtype.kind == 'f'
    return np.einsum('hwc,h,w -> hwc', tile, weight, weight, optimize=True)


# ------------------------------ mosaic setting ------------------------------


@dataclass(frozen=True, slots=True)
class Mosaic:
    """
    Helper to split image to tiles and process them.

    Parameters:
    - step - Step between consecutive tiles
    - overlap - Count of pixels that will be shared among overlapping tiles

    So tile size = overlap + non-overlapped area + overlap = step + overlap,
    and non-overlapped area = step - overlap.
    """
    step: int
    overlap: int

    def __post_init__(self):
        if self.overlap < 0:
            raise ValueError('overlap should be non-negative, '
                             f'got: {self.overlap}')
        # That may be optional
        if self.overlap % 2 != 0:
            raise ValueError(f'overlap should be even, got: {self.overlap}')
        if self.overlap > self.step:
            raise ValueError('overlap should be lower than step, got: '
                             f'overlap={self.overlap} and step={self.step}')

    def get_kernel(self) -> np.ndarray:
        return get_trapz(self.step, self.overlap)

    def iterate(self,
                image: NumpyLike,
                max_workers: int = 1) -> '_TiledArrayView':
        """Read tiles from input image"""
        shape = image.shape[:2]
        ishape = *(len(range(0, s + self.overlap, self.step)) for s in shape),
        cells = np.ones(ishape, dtype=np.bool_)

        return _TiledArrayView(self, shape, cells, image, max_workers)


# --------------------------------- actions ----------------------------------

_Self = TypeVar('_Self', bound='_BaseView')


@dataclass(frozen=True)
class _BaseView:
    m: Mosaic
    shape: tuple[int, ...]
    cells: np.ndarray

    @property
    def ishape(self) -> tuple[int, ...]:
        return self.cells.shape

    def __len__(self) -> int:
        return int(self.cells.sum())

    def report(self) -> dict[str, str]:
        """Cells and area usage"""
        used = int(self.cells.sum())
        total = self.cells.size
        coverage = (used / total) * (1 + self.m.overlap / self.m.step) ** 2
        return {'cells': f'{used}/{total}', 'coverage': f'{coverage:.0%}'}

    def __iter__(self) -> Iterator[Tile]:
        raise NotImplementedError

    def map(self: _Self,
            fn: Callable[[np.ndarray], np.ndarray],
            /,
            max_workers: int = 0) -> _Self:
        """
        Applies function to each tile.
        Note: change of tile shape besides channel count is forbidden.
        Each tile is HWC-ordered ndarray.
        Supports threading.
        """
        tile_fn = partial(_apply, fn)
        tiles = map_n(tile_fn, self, max_workers=max_workers)
        return cast(
            _Self,  # don't narrow type
            _IterView(self.m, self.shape, self.cells, tiles),
        )

    def pool(self: _Self, stride: int = 1) -> _Self:
        """Resizes each tile to desired stride"""
        if stride == 1:
            return self
        if self.m.step % stride != 0 or self.m.overlap % stride != 0:
            raise ValueError('stride should be a divisor of '
                             f'{self.m.step}, {self.m.overlap}. Got: {stride}')

        m = Mosaic(self.m.step // stride, self.m.overlap // stride)
        shape = *(len(range(0, s, stride)) for s in self.shape),
        return cast(  # don't narrow type
            _Self,
            _DecimatedView(m, shape, self.cells, stride, self),
        )

    def crop(self) -> '_BaseView':
        it = map(Cropper(self.shape), self)
        return _IterView(self.m, self.shape, self.cells, it)

    def zip_with(self, view: np.ndarray, v_scale: float) -> '_ZipView':
        """Extracts tiles from `view` simultaneously with tiles from self"""
        if v_scale > 1:
            raise ValueError('v_scale should be less than 1, '
                             f'got: {v_scale}')
        return _ZipView(self, view, v_scale)

    def with_cm(self: _Self, ctx: AbstractContextManager) -> _Self:
        return cast(
            _Self,
            _ScopedView(self.m, self.shape, self.cells, self, ctx),
        )


@dataclass(frozen=True)
class _ZipView:
    source: _BaseView
    view: NumpyLike
    v_scale: float

    def __len__(self) -> int:
        return len(self.source)

    def __iter__(self) -> Iterator[tuple[Vec, Vec, np.ndarray, np.ndarray]]:
        return map(Zipper(self.view, self.v_scale), self.source)


@dataclass(frozen=True)
class _View(_BaseView):
    def map_batched(self,
                    fn: Callable[[list[np.ndarray]], Iterable[np.ndarray]],
                    /,
                    batch_size: int = 1,
                    max_workers: int = 0) -> '_View':
        """
        Applies function to batches of tiles.
        Note: change of tile shape besides channel count is forbidden.
        Each tile is HWC-ordered ndarray.
        Supports threading.
        """
        tile_fn = partial(_apply_batched, fn)

        chunks = chunked(self, batch_size)
        batches = map_n(tile_fn, chunks, max_workers=max_workers)
        tiles = chain.from_iterable(batches)

        return _IterView(self.m, self.shape, self.cells, tiles)

    def transform(self, fn: Callable[[Iterable[Tile]],
                                     Iterable[Tile]]) -> '_View':
        """
        TODO: fill docs
        """
        tiles = fn(self)
        return _IterView(self.m, self.shape, self.cells, tiles)

    def reweight(self) -> '_View':
        """
        Applies weight to tile edges to prepare them for summation
        if overlap exists.
        Note: no need to call this method if you have already applied it.
        """
        if not self.m.overlap:  # No need
            return self

        weight = self.m.get_kernel()
        tile_fn = partial(_reweight, weight)
        return self.map(tile_fn)

    def merge(self) -> _BaseView:
        """
        Removes overlapping regions from all the tiles if any.
        Tiles should be reweighted if overlap exists.
        """
        if self.m.overlap:
            return _BlendCropsView(self.m, self.shape, self.cells, self)
        return self


@dataclass(frozen=True)
class _IterView(_View):
    source: Iterable[Tile]

    def __iter__(self) -> Iterator[Tile]:
        return iter(self.source)


@dataclass(frozen=True)
class _ScopedView(_BaseView):
    source: Iterable[Tile]
    _ctx: AbstractContextManager

    def __iter__(self) -> Iterator[Tile]:
        """Iterator over tiles. Reentrant if nested context is reentrant"""
        with self._ctx:
            yield from self.source


@dataclass(frozen=True)
class _DecimatedView(_View):
    """
    Decimates tiles.
    Doesn't change size uniformity.
    Yields decimated views of original tiles.
    """
    stride: int
    source: Iterable[Tile]

    def __iter__(self) -> Iterator[Tile]:
        return map(Decimator(self.stride), self.source)


@dataclass(frozen=True)
class _TiledArrayView(_View):
    """
    Extracts tiles from array.
    Yields same-sized tiles with overlaps.
    """
    data: NumpyLike
    max_workers: int

    def select(self,
               mask: np.ndarray,
               scale: float | None = None) -> '_TiledArrayView':
        """
        Subset non-masked (i.e. non-zeros in mask) tiles for iteration.
        Mask size is "scale"-multiple of source image.
        """
        if mask.ndim == 3:  # Strip extra dim
            mask = mask.squeeze(2)
        if mask.ndim != 2:
            raise ValueError(f'Mask should be 2D, got shape: {mask.shape}')
        if scale is None:
            scale = max(ms / s for ms, s in zip(mask.shape, self.data.shape))

        mask = np.where(mask, np.uint8(255), np.uint8(0))

        tile = (self.m.step + self.m.overlap) * scale
        if (h_tile := ceil(tile / 2)):
            kernel = np.ones((3, 3), dtype='u1')
            mask = cv2.dilate(mask, kernel, iterations=h_tile)

        ih, iw = self.ishape
        cells = cv2.resize(mask, (iw, ih), interpolation=cv2.INTER_AREA)

        return dataclasses.replace(self, cells=cells.astype(bool))

    def _slice_tile(self, iy: int, ix: int) -> Tile:
        """Slice source image to get overlapping tiles"""
        (y0, y1), (x0, x1) = ((
            self.m.step * i - self.m.overlap,
            self.m.step * (i + 1),
        ) for i in (iy, ix))

        data = padslice(self.data, slice(y0, y1), slice(x0, x1))
        return Tile(idx=(iy, ix), vec=(y0, x0), data=data)

    def _get_tile(self, iy: int, ix: int) -> Tile:
        """Read non-overlapping tile of source image"""
        # First tile is [-overlap: step], as `step + overlap = tile`
        (y0, y1), (x0, x1) = ((
            self.m.step * i - self.m.overlap,
            self.m.step * (i + 1),
        ) for i in (iy, ix))

        # We have tile above to pick upper edge from
        if iy and self.cells[iy - 1, ix]:
            y0 += self.m.overlap
        # We have previous tile to pick left edge from
        if ix and self.cells[iy, ix - 1]:
            x0 += self.m.overlap

        return Tile(
            idx=(iy, ix),
            vec=(y0, x0),
            data=self.data[y0:y1, x0:x1],
        )

    def __iter__(self) -> Iterator[Tile]:
        """
        Yield complete tiles built from source image.
        Each tile will have size `(step + overlap)`
        """
        ys, xs = np.where(self.cells)
        if isinstance(self.data, np.ndarray):
            return map(self._slice_tile, ys, xs)

        parts = map_n(self._get_tile, ys, xs, max_workers=self.max_workers)
        if not self.m.overlap:
            return iter(parts)

        rcr = Reconstructor(self.m.overlap, self.m.step, self.cells)
        return map(rcr, parts)

    def get_tile(self, idx: int) -> Tile:
        iy, ix = np.argwhere(self.cells)[idx]
        return self._slice_tile(iy, ix)


@dataclass(frozen=True)
class _BlendCropsView(_BaseView):
    """
    Applies weighted average over overlapping regions.
    Yields tiles without overlaps, so their size can differ.
    """
    source: Iterable[Tile]

    def __iter__(self) -> Iterator[Tile]:
        assert self.m.overlap
        bcr = BlendCropper(self.m.step, self.m.overlap, self.cells)
        return map(bcr, self.source)



__all__ = [
    'get_fusion', 'get_trapz', 'normalize_loc', 'probs_to_rgb_heatmap',
    'resize'
]

from collections.abc import Iterable, Sequence
from itertools import zip_longest

import cv2
import numpy as np

from bipl import env

from ._types import NumpyLike, Tile


def probs_to_rgb_heatmap(prob: np.ndarray) -> np.ndarray:
    """Converts probability array of (H W C) shape to (H W 3) RGB image"""
    if prob.ndim != 3:
        raise ValueError(f'prob should be 3d, got: {prob.shape}')
    h, w, c = prob.shape
    if not prob.size:
        return np.empty((h, w, 3), dtype='u1')

    hue = prob.argmax(-1).astype('u1')
    value = prob.max(-1)
    if value.dtype != 'u1':
        value *= 255
        value = value.round().astype('u1')

    hue_lut = np.zeros(256, 'u1')
    hue_lut[:c] = np.linspace(0, 127, c, endpoint=False, dtype='u1')
    hsv = cv2.merge([
        cv2.LUT(hue, hue_lut),
        np.broadcast_to(np.uint8(255), (h, w)),
        value,
    ])
    return cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)


def get_trapz(step: int, overlap: int) -> np.ndarray:
    """Returns trapezional window kernel to apply stiching"""
    if overlap == 0:
        raise ValueError('not applicable for overlap 0')
    pad = np.linspace(0, 1, overlap + 2)[1:-1]  # strip initial 0 and final 1
    return np.r_[pad, np.ones(step - overlap), pad[::-1]].astype('f4')


def normalize_loc(loc: Sequence[slice] | slice,
                  shape: Sequence[int]) -> tuple[slice, ...]:
    """Ensures slices match shape and have non-none endpoints"""
    if isinstance(loc, slice):
        loc = loc,
    ndim = len(shape)
    if len(loc) > ndim:
        raise ValueError(f'loc is too deep for {ndim}D, got: {loc}')
    it = (slice(
        s.start or 0,
        s.stop if s.stop is not None else axis_len,
        s.step if s.step is not None else 1,
    ) for s, axis_len in zip_longest(loc, shape, fillvalue=slice(None)))
    return *it,


def padslice(a: NumpyLike, *loc: slice) -> np.ndarray:
    """Do `a[loc]`, but extend `a` if `loc` indices beyond `a.shape`"""
    loc = normalize_loc(loc, a.shape)

    pos_loc = *(slice(max(s.start, 0), s.stop) for s in loc),
    a = a[pos_loc]

    pad = [(pos.start - raw.start, pos.stop - pos.start - size)
           for raw, pos, size in zip(loc, pos_loc, a.shape)]
    return np.pad(a, pad) if np.any(pad) else a


def resize(image: np.ndarray,
           hw: tuple[int, ...],
           *,
           antialias: bool | None = None) -> np.ndarray:
    if image.shape[:2] == hw[:2]:
        return image

    f_max = max(s0 / s1 for s0, s1 in zip(image.shape, hw))
    if f_max < 1:  # Upscaling
        antialias = False
    elif antialias is None:  # Enable only if we downscale more than 2.5x
        antialias = f_max > 2.5

    h, w = hw
    interpolation = cv2.INTER_LINEAR
    if antialias:  # Downsampling
        match env.BIPL_DOWN:
            case 'box2d':  # Pyramid, box filter. Fastest
                for _ in range(int(f_max).bit_length() - 1):
                    image = cv2.resize(image, None, fx=0.5, fy=0.5)

            case 'gauss':  # Pyramid, gauss filter. No aliasing. 2X slower
                for _ in range(int(f_max).bit_length() - 1):
                    image = cv2.pyrDown(image)

            case 'area':  # No pyramid, box filter. Slowest
                interpolation = cv2.INTER_AREA

    return cv2.resize(image, (w, h), interpolation=interpolation)


def get_fusion(tiles: Iterable[Tile],
               shape: tuple[int, ...] | None = None) -> np.ndarray | None:
    r: np.ndarray | None = None

    if shape is None:  # Collect all the tiles to compute destination size
        tiles = [*tiles]
        if not tiles:
            return None
        # N arrays of (yx + hw)
        yx_hw = np.array([[[t.vec, t.data.shape[:2]] for t in tiles]]).sum(1)
        # bottom left most edge
        shape = *yx_hw.max(0).tolist(),
    elif len(shape) != 2:
        raise ValueError(f'shape should be 2-tuple, got {shape}')

    for _, (y, x), tile in tiles:
        if not tile.size:
            continue

        h, w, c = tile.shape
        if r is None:  # First iteration, initilize
            r = np.zeros((*shape, c), tile.dtype)

        if c != r.shape[2]:
            raise RuntimeError('tile channel counts changed during iteration')
        v = r[y:, x:][:h, :w]  # View to destination
        v[:] = tile[:v.shape[0], :v.shape[1]]  # Crop if needed

    return r
